from abc import ABC, abstractmethod
from functools import partial
from typing import Any, Callable, Self

from pyjudo.service_life import ServiceLife


class IServiceContainer(ABC):
    @abstractmethod
    def register[T](
        self,
        abstract_class: type[T],
        constructor: type[T] | Callable[..., T],
        service_life: ServiceLife = ServiceLife.TRANSIENT,
    ) -> Self: ...

    @abstractmethod
    def unregister(self, abstract_class: type) -> Self: ...

    @abstractmethod
    def add_transient[T](self, abstract_class: type[T], service_class: type[T]) -> Self: ...

    @abstractmethod
    def add_scoped[T](self, abstract_class: type[T], service_class: type[T]) -> Self: ...

    @abstractmethod
    def add_singleton[T](self, abstract_class: type[T], service_class: type[T]) -> Self: ...

    @abstractmethod
    def get[T](self, abstract_class: type[T], **overrides: Any) -> T: ... # pyright: ignore[reportAny]

    @abstractmethod
    def is_registered(self, abstract_class: type) -> bool: ...

    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        return partial(self.get, key)


from collections.abc import Callable
import pytest
from abc import ABC

from pyjudo import ServiceContainer, Factory

from pyjudo.exceptions import (
    ServiceRegistrationError,
    ServiceResolutionError,
    ServiceTypeError,
)


def test_factory_registration(services):
    container = ServiceContainer()

    def factory() -> services.IServiceA:
        return services.ServiceA("Factory")

    _ = container.register(services.IServiceA, factory)

    assert container.is_registered(services.IServiceA)

    instance = container.get(services.IServiceA)
    assert instance.value == "Factory"


def test_factory_registration_with_dependencies(services):
    container = ServiceContainer()

    class IAnotherService(ABC): ...

    class AnotherService(IAnotherService): ...

    def factory(another: IAnotherService) -> services.IServiceA:
        return services.ServiceA()

    _ = container.register(IAnotherService, AnotherService)
    _ = container.register(services.IServiceA, factory)

    instance = container.get(services.IServiceA)
    assert instance.value == "A"


def test_factory_registration_with_missing_dependencies(services):
    container = ServiceContainer()

    class IAnotherService(ABC): ...

    class AnotherService(IAnotherService): ...

    def factory(another: IAnotherService) -> services.IServiceA:
        return services.ServiceA()

    _ = container.register(services.IServiceA, factory)

    with pytest.raises(ServiceResolutionError):
        instance = container.get(services.IServiceA)


def test_factory_in_dependencies(services):
    container = ServiceContainer()

    class IAnotherService(ABC):
        factory: Callable[..., services.IServiceA]

    class AnotherService(IAnotherService):
        def __init__(self, factory: Factory[services.IServiceA]):
            self.factory = factory

    _ = container.register(services.IServiceA, services.ServiceA)
    _ = container.register(IAnotherService, AnotherService)

    instance = container.get(IAnotherService)

    assert callable(instance.factory)
    assert instance.factory.__repr__().startswith("FactoryProxy")

    service_a = instance.factory()

    assert isinstance(service_a, services.ServiceA)

def test_factory_registration_without_return_annotation(services):
    container = ServiceContainer()

    def factory():
        return services.ServiceA()

    with pytest.raises(ServiceRegistrationError):
        _ = container.register(services.IServiceA, factory)

def test_factory_registration_with_incorrect_return_annotation(services):
    container = ServiceContainer()

    def factory() -> services.IServiceB:
        return services.ServiceA()

    with pytest.raises(ServiceRegistrationError):
        _ = container.register(services.IServiceA, factory)
    

def test_factory_function_with_wrong_return_type(services):
    container = ServiceContainer()

    def factory() -> services.IServiceA:
        return "nonsense"

    _ = container.register(services.IServiceA, factory)

    with pytest.raises(ServiceTypeError):
        instance = container.get(services.IServiceA)


def test_factory_registration_not_class_or_callable(services):
    container = ServiceContainer()

    with pytest.raises(ServiceRegistrationError):
        _ = container.register(services.IServiceA, "Not a class or callable")

def test_get_factory(services):
    container = ServiceContainer()

    _ = container.register(services.IServiceA, services.ServiceA)

    factory = container.get_factory(services.IServiceA)

    assert callable(factory)

    instance = factory()

    assert isinstance(instance, services.ServiceA)

def test_get_factory_not_registered(services):
    container = ServiceContainer()

    with pytest.raises(ServiceResolutionError):
        factory = container.get_factory(services.IServiceA)


import pytest

from pyjudo import ServiceContainer, ServiceLife
from pyjudo.exceptions import ServiceResolutionError, ServiceScopeError, ServiceDisposedError


def test_scoped_lifetime(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.SCOPED)

    with container.create_scope() as scope:
        instance1 = scope.get(services.IServiceA)
        instance2 = scope.get(services.IServiceA)

        assert instance1 is instance2
        assert instance1.value == "A"

    with container.create_scope():
        instance3 = container.get(services.IServiceA)
        instance4 = container.get(services.IServiceA)

        assert instance3 is instance4
        assert instance3.value == "A"

    with container.create_scope() as scope:
        service_a_partial = scope[services.IServiceA]

        assert callable(service_a_partial)
        instance5 = service_a_partial()
        assert isinstance(instance5, services.ServiceA)

def test_scoped_lifetime_multiple_scopes(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.SCOPED)

    with container.create_scope() as scope1:
        instance1 = scope1.get(services.IServiceA)

        with container.create_scope() as scope2:
            instance2 = scope2.get(services.IServiceA)

            assert instance1 is not instance2
            assert instance1.value == "A"
            assert instance2.value == "A"


def test_scoped_with_no_scope(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.SCOPED)

    with pytest.raises(ServiceResolutionError):
        _ = container.get(services.IServiceA)


def test_scoped_with_disposable(services):
    container = ServiceContainer()
    _ = container.register(
        services.IServiceA, services.SoftDisposableService, ServiceLife.SCOPED
    )

    with container.create_scope() as scope:
        instance1 = scope.get(services.IServiceA)

    assert instance1.value == "disposed"


def test_scoped_pop_exception(services):
    container = ServiceContainer()
    
    scope = container.create_scope()
    container._push_scope(scope)
    container._pop_scope()
    
    with pytest.raises(ServiceScopeError):
        container._pop_scope()

def test_scoped_disposable(services):
    container = ServiceContainer()
    _ = container.register(
        services.IServiceA, services.HardDisposableService, ServiceLife.SCOPED
    )

    with container.create_scope() as scope:
        instance = scope.get(services.IServiceA)

    assert instance.is_disposed

    with pytest.raises(ServiceDisposedError):
        _ = instance.value




from abc import ABC
from types import SimpleNamespace
import pytest

from pyjudo.disposable import IDisposable

# Mock services for testing
class IServiceA(ABC):
    value: str

class IServiceB(ABC):
    service_a: IServiceA

class IServiceC(ABC):
    service_a: IServiceA
    service_b: IServiceB


class ServiceA(IServiceA):
    def __init__(self, value: str = "A"):
        self.value: str = value


class ServiceB(IServiceB):
    def __init__(self, service_a: IServiceA):
        self.service_a: IServiceA = service_a
        self.value: str = "B"


class ServiceC(IServiceC):
    def __init__(self, service_b: IServiceB, service_a: IServiceA):
        self.service_b: IServiceB = service_b
        self.service_a: IServiceA = service_a
        self.value: str = "C"


class CircularService(IServiceA):
    def __init__(self, service_c: IServiceC):
        self.service_c: IServiceC = service_c


class SoftDisposableService(IServiceA):
    def __init__(self):
        self.value = "A"

    def dispose(self):
        self.value = "disposed"


class HardDisposableService(IServiceA, IDisposable):
    def __init__(self):
        self.value = "A"

    def do_dispose(self):
        self.value = "disposed"


@pytest.fixture
def services():
    return SimpleNamespace(
        IServiceA=IServiceA,
        IServiceB=IServiceB,
        IServiceC=IServiceC,
        ServiceA=ServiceA,
        ServiceB=ServiceB,
        ServiceC=ServiceC,
        CircularService=CircularService,
        SoftDisposableService=SoftDisposableService,
        HardDisposableService=HardDisposableService,
    )


import pytest

from pyjudo import ServiceContainer
from pyjudo.exceptions import (
    ServiceCircularDependencyError,
    ServiceResolutionError,
    ServiceRegistrationError,
)


def test_register_service(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA)

    assert container.is_registered(services.IServiceA)

    container = ServiceContainer()
    _ = container.add_transient(services.IServiceA, services.ServiceA)

    assert container.is_registered(services.IServiceA)

    container = ServiceContainer()
    _ = container.add_scoped(services.IServiceA, services.ServiceA)

    assert container.is_registered(services.IServiceA)

    container = ServiceContainer()
    _ = container.add_singleton(services.IServiceA, services.ServiceA)

    assert container.is_registered(services.IServiceA)


def test_get_resolved_service(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA)
    _ = container.register(services.IServiceB, services.ServiceB)

    service_b = container.get(services.IServiceB)

    assert isinstance(service_b, services.ServiceB)
    assert isinstance(service_b.service_a, services.ServiceA)
    assert service_b.value == "B"

    service_a_partial = container[services.IServiceA]

    assert callable(service_a_partial)
    service_a = service_a_partial()
    assert isinstance(service_a, services.ServiceA)


def test_circular_dependency_detection(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.CircularService)
    _ = container.register(services.IServiceB, services.ServiceB)
    _ = container.register(services.IServiceC, services.ServiceC)

    with pytest.raises(ServiceCircularDependencyError):
        _ = container.get(services.IServiceA)


def test_unregistered_service_resolution(services):
    container = ServiceContainer()

    with pytest.raises(ServiceResolutionError):
        _ = container.get(services.IServiceA)


def test_duplicate_registration(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA)

    with pytest.raises(ServiceRegistrationError):
        _ = container.register(services.IServiceA, services.ServiceA)

def test_registration_incorrect_type(services):
    container = ServiceContainer()

    with pytest.raises(ServiceRegistrationError):
        _ = container.register(services.IServiceA, services.ServiceB)

def test_unregister(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA)

    container.unregister(services.IServiceA)

    assert not container.is_registered(services.IServiceA)

def test_unregister_unregistered_service(services):
    container = ServiceContainer()

    with pytest.raises(ServiceRegistrationError):
        container.unregister(services.IServiceA)


class ServiceException(Exception):
    pass


class ServiceCircularDependencyError(ServiceException):
    pass


class ServiceResolutionError(ServiceException):
    pass


class ServiceRegistrationError(ServiceException):
    pass


class ServiceTypeError(ServiceException):
    pass


class ServiceDisposedError(ServiceException):
    pass

class ServiceScopeError(ServiceException):
    pass


from abc import ABC, abstractmethod
from typing import Any, Protocol, override, runtime_checkable

from pyjudo.exceptions import ServiceDisposedError

@runtime_checkable
class Disposable(Protocol):
    def dispose(self) -> None: ...


class IDisposable(ABC):
    """
    Represents an object that can be disposed.
    """
    __disposed: bool = False

    @abstractmethod
    def do_dispose(self) -> None:
        """
        Custom dispose logic should be implemented in subclasses.
        """
        pass

    def dispose(self) -> None:
        """
        Disposes of the object, setting its disposed state to True and calling the custom dispose logic.
        """
        if not self.__disposed:
            self.do_dispose()
            self.__disposed = True

    @property
    def is_disposed(self) -> bool:
        """
        Returns whether the object has been disposed.
        """
        return self.__disposed

    def _check_disposed(self) -> None:
        """
        Raises an exception if the object has been disposed.
        """
        if self.__disposed:
            raise ServiceDisposedError("Object is disposed and cannot be used.")

    def __getattr__(self, name: str): # pyright: ignore[reportAny]
        """
        Checks if the object is disposed before accessing any attributes.
        """
        self._check_disposed()
        return super().__getattribute__(name) # pyright: ignore[reportAny]

    @override
    def __getattribute__(self, name: str): # pyright: ignore[reportAny]
        """
        Checks if the object is disposed before accessing any attributes.
        """
        # Allow access to certain attributes to prevent recursion and allow checking if disposed.
        permitted_attrs = (
            "is_disposed",
            "_check_disposed", 
            "_IDisposable__disposed")
        if name not in permitted_attrs:
            self._check_disposed()
        return super().__getattribute__(name) # pyright: ignore[reportAny]

    @override
    def __setattr__(self, name: str, value: Any) -> None: # pyright: ignore[reportAny]
        """
        Checks if the object is disposed before setting any attributes.
        """
        self._check_disposed()
        super().__setattr__(name, value)

    def __call__(self, *args: Any, **kwargs: Any) -> Any: # pyright: ignore[reportAny]
        """
        Checks if the object is disposed before calling it.
        """
        self._check_disposed()
        return super().__call__(*args, **kwargs) # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportAttributeAccessIssue]

    @override
    def __str__(self) -> str:
        """
        Checks if the object is disposed before converting it to a string.
        """
        self._check_disposed()
        return f"DISPOSED<{super().__str__()}>" if self.__disposed else super().__str__()

    @override
    def __repr__(self) -> str:
        """
        Checks if the object is disposed before getting its representation.
        """
        return f"DISPOSED<{super().__repr__()}>" if self.__disposed else super().__repr__()

    def __del__(self) -> None:
        """
        Ensures dispose is called before garbage collection if not already done.
        """
        if not self.__disposed:
            self.dispose()


from collections.abc import Callable
import threading

from pyjudo.service_life import ServiceLife



class ServiceEntry[T]:
    """
    Represents a service entry in the container.
    """
    constructor: type[T] | Callable[..., T]
    service_life: ServiceLife
    instance: T | None
    lock: threading.Lock

    def __init__(self, constructor: type[T] | Callable[..., T], service_life: ServiceLife):
        self.constructor = constructor
        self.service_life = service_life
        self.instance = None
        self.lock = threading.Lock()


import logging
import inspect
import threading
from typing import Any, Callable, cast, get_args, get_origin, override, Self
from functools import partial

from pyjudo.exceptions import (
    ServiceCircularDependencyError,
    ServiceResolutionError,
    ServiceRegistrationError,
    ServiceTypeError,
    ServiceScopeError,
)
from pyjudo.factory import Factory, FactoryProxy
from pyjudo.iservice_container import IServiceContainer
from pyjudo.service_entry import ServiceEntry
from pyjudo.service_life import ServiceLife
from pyjudo.service_scope import ServiceScope


class ServiceContainer(IServiceContainer):
    """
    A container for registering and resolving services with dependency injection.
    """

    def __init__(self):
        self.__lock = threading.Lock()
        self.__resolution_stack = threading.local()
        self.__scopes_stack = threading.local()

        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self._services: dict[type[Any], ServiceEntry[Any]] = {}

        this_service_entry = ServiceEntry[Self](
            self.__class__,
            ServiceLife.SINGLETON,
        )
        this_service_entry.instance = self

        self._services[IServiceContainer] = this_service_entry
        self._logger.debug("Initialised service container")

    def _set_service[T](self, key: type[T], value: ServiceEntry[T]) -> None:
        with self.__lock:
            self._services[key] = value
    
    def _get_service[T](self, key: type[T]) -> ServiceEntry[T]:
        with self.__lock:
            try:
                return self._services[key]
            except KeyError:
                raise ServiceResolutionError(f"Unable to find service: {key}")

    @property
    def _scope_stack(self) -> list[ServiceScope]:
        if not hasattr(self.__scopes_stack, "scopes"):
            self.__scopes_stack.scopes = []
        return self.__scopes_stack.scopes

    @property
    def _resolution_stack(self) -> set[type]:
        if not hasattr(self.__resolution_stack, "stack"):
            self.__resolution_stack.stack = set()
            self._logger.debug("Initialized a new resolution stack for the thread.")
        return self.__resolution_stack.stack

    def _current_scope(self) -> ServiceScope | None:
        if not self._scope_stack:
            return None
        return self._scope_stack[-1]

    def _push_scope(self, scope: ServiceScope) -> None:
        with self.__lock:
            self._scope_stack.append(scope)
            self._logger.debug("Pushed new scope to stack.")

    def _pop_scope(self) -> None:
        with self.__lock:
            try:
                _ = self._scope_stack.pop()
                self._logger.debug("Popped scope from stack.")
            except IndexError:
                raise ServiceScopeError("No scope available to pop.")

    @override
    def register[T](
        self,
        abstract_class: type[T],
        constructor: type[T] | Callable[..., T],
        service_life: ServiceLife = ServiceLife.TRANSIENT,
    ) -> Self:
        """
        Registers a service class with the container.

        :param abstract_class: The abstract class or interface.
        :param service_class: The concrete implementation of abstract_class.
        :param service_life: The lifecycle of the service.
        """
        if abstract_class in self._services:
            raise ServiceRegistrationError(
                f"Service '{abstract_class.__name__}' is already registered."
            )

        if inspect.isclass(constructor):
            if not issubclass(constructor, abstract_class):
                raise ServiceRegistrationError(f"'{constructor.__name__}' does not implement '{abstract_class.__name__}'")
        elif callable(constructor):
            return_annotation = inspect.signature(constructor).return_annotation

            if return_annotation is inspect.Signature.empty:
                raise ServiceRegistrationError(f"Callable '{constructor.__name__}' must have a return annotation.")
            
            if not issubclass(return_annotation, abstract_class):
                raise ServiceRegistrationError(f"'{constructor.__name__}' does not return '{abstract_class.__name__}'")
        else:
            raise ServiceRegistrationError("Constructor must be a class or callable")

        service = ServiceEntry[T](constructor, service_life)
        self._set_service(abstract_class, service)
        self._logger.debug(f"Registered service: {abstract_class.__name__} as {constructor.__name__} with life {service_life.name}")
        return self

    @override
    def unregister(self, abstract_class: type) -> Self:
        with self.__lock:
            try:
                del self._services[abstract_class]
                self._logger.debug(f"Unregistered service: {abstract_class.__name__}")
            except KeyError:
                raise ServiceRegistrationError(f"Service '{abstract_class.__name__}' is not registered.")
        return self

    @override
    def add_transient[T](self, abstract_class: type[T], service_class: type[T]) -> Self:
        return self.register(abstract_class, service_class, ServiceLife.TRANSIENT)
    
    @override
    def add_scoped[T](self, abstract_class: type[T], service_class: type[T]) -> Self:
        return self.register(abstract_class, service_class, ServiceLife.SCOPED)
    
    @override
    def add_singleton[T](self, abstract_class: type[T], service_class: type[T]) -> Self:
        return self.register(abstract_class, service_class, ServiceLife.SINGLETON)

    @override
    def get[T](self, abstract_class: type[T], **overrides: Any) -> T: # pyright: ignore[reportAny]
        service = self._resolve(abstract_class, scope=self._current_scope(), **overrides)
        if not issubclass(type(service), abstract_class):
            raise ServiceTypeError(f"Service '{service}' is not of type '{abstract_class.__name__}'")
        return service

    def get_factory[T](self, abstract_class: type[T]) -> Callable[..., T]:
        if not self.is_registered(abstract_class):
            raise ServiceResolutionError(f"Service '{abstract_class.__name__}' is not registered.")
        return FactoryProxy(self, abstract_class)

    def _resolve[T](self, abstract_class: type[T], scope: ServiceScope | None, **overrides: Any) -> T:
        if abstract_class in self._resolution_stack:
            raise ServiceCircularDependencyError(
                f"Circular dependency detected for '{abstract_class.__name__}'"
            )

        _ = self._resolution_stack.add(abstract_class)
        self._logger.debug(f"Resolving service '{abstract_class.__name__}'")

        try:
            service = self._get_service(abstract_class)

            match service.service_life:
                case ServiceLife.SINGLETON:
                    return self._get_singleton(service, **overrides)
                case ServiceLife.SCOPED:
                    if scope is None:
                        raise ServiceResolutionError(
                            f"Service '{abstract_class.__name__}' is scoped but no scope was provided."
                        )
                    return self._get_scoped(abstract_class, scope, service, **overrides)
                case ServiceLife.TRANSIENT:
                    return self._get_transient(service, **overrides)
        finally:
            self._resolution_stack.remove(abstract_class)

    @override
    def is_registered(self, abstract_class: type) -> bool:
        return abstract_class in self._services

    def _get_singleton[T](self, service_entry: ServiceEntry[T], **overrides: Any) -> T:
        if service_entry.instance is None:
            service_entry.instance = self._create_instance(service_entry, overrides)
        else:
            if overrides:
                raise ServiceResolutionError("Cannot use overrides with a singleton which has already been resolved.")
        return service_entry.instance

    def _get_scoped[T](self, abstract_class: type[T], scope: ServiceScope, service_entry: ServiceEntry[T], **overrides: Any) -> T:
        if scope.has_instance(abstract_class):
            return scope.get_instance(abstract_class)
        instance = self._create_instance(service_entry, overrides)
        scope.set_instance(abstract_class, instance)
        return instance

    def _get_transient[T](self, service_entry: ServiceEntry[T], **overrides: Any) -> T:
        return self._create_instance(service_entry, overrides)

    def _create_instance[T](self, service_entry: ServiceEntry[T], overrides: dict[str, Any]) -> T:
        if inspect.isclass(service_entry.constructor):
            type_hints = inspect.signature(service_entry.constructor.__init__).parameters
        else:
            type_hints = inspect.signature(service_entry.constructor).parameters
        
        kwargs = {}
        
        for name, param in type_hints.items():
            if name == "self":
                continue

            # Skip *args and **kwargs
            if param.kind in (
                inspect.Parameter.VAR_POSITIONAL,
                inspect.Parameter.VAR_KEYWORD,
            ):
                continue
            
            origin = get_origin(param.annotation)
            args = get_args(param.annotation)
            if origin is Factory and args:
                abstract_class = args[0]
                kwargs[name] = FactoryProxy(self, abstract_class)
                continue

            if name in overrides:
                kwargs[name] = overrides[name]
            elif param.annotation in self._services:
                kwargs[name] = self.get(param.annotation)
            elif param.default != inspect.Parameter.empty:
                kwargs[name] = param.default
            else:
                raise ServiceResolutionError(f"Unable to resolve dependency '{name}' for '{service_entry.constructor}'")
        
        self._logger.debug(f"Creating new instance of '{T}'")

        return cast(T, service_entry.constructor(**kwargs)) # TODO: Remove cast

    def create_scope(self) -> ServiceScope:
        return ServiceScope(self)

    @override
    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        return partial(self.get, key)


import pytest

from pyjudo import ServiceContainer, ServiceLife
from pyjudo.exceptions import ServiceResolutionError


def test_singleton_lifetime(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.SINGLETON)

    instance1 = container.get(services.IServiceA)
    instance2 = container.get(services.IServiceA)

    assert instance1 is instance2
    assert instance1.value == "A"

def test_overrides_in_singleton(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.SINGLETON)

    # First instantiation without overrides
    instance1 = container.get(services.IServiceA)
    assert instance1.value == "A"

    # Singleton should prevent overrides after the instance is created
    with pytest.raises(ServiceResolutionError):
        _ = container.get(services.IServiceA, value="Should Fail")



from pyjudo import ServiceContainer, ServiceLife


def test_transient_lifetime(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.TRANSIENT)

    instance1 = container.get(services.IServiceA)
    instance2 = container.get(services.IServiceA)

    assert instance1 is not instance2
    assert instance1.value == "A"
    assert instance2.value == "A"


def test_overrides_in_transient(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.TRANSIENT)

    # Overriding 'value' attribute for transient instance
    service_a = container.get(services.IServiceA, value="Overridden")

    assert service_a.value == "Overridden"

def test_transient_dependencies(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.TRANSIENT)
    _ = container.register(services.IServiceB, services.ServiceB, ServiceLife.TRANSIENT)
    _ = container.register(services.IServiceC, services.ServiceC, ServiceLife.TRANSIENT)

    service_b = container.get(services.IServiceB)
    service_c = container.get(services.IServiceC)

    assert service_b.service_a is not service_c.service_a


from functools import partial
import logging
from typing import Any, Callable, TYPE_CHECKING

from pyjudo.disposable import Disposable

if TYPE_CHECKING:
    from pyjudo.service_container import ServiceContainer


class ServiceScope:
    """
    Represents a scope for services.
    """
    def __init__(self, service_container: "ServiceContainer") -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self._instances: dict[type[Any], Any] = {}
        self._disposables: list[Disposable] = []
        self._container: "ServiceContainer" = service_container
    
    def get[T](self, abstract_class: type[T], **overrides: Any) -> T:
        return self._container._resolve(abstract_class, scope=self, overrides=overrides)

    def has_instance(self, abstract_class: type) -> bool:
        return abstract_class in self._instances
    
    def get_instance[T](self, abstract_class: type[T]) -> T:
        return self._instances[abstract_class]

    def set_instance(self, abstract_class: type, instance: Any) -> None:
        self._instances[abstract_class] = instance
        if isinstance(instance, Disposable):
            self._disposables.append(instance)
    
    def __enter__(self):
        self._container._push_scope(self)
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        for instance in self._disposables:
            instance.dispose()
        self._container._pop_scope()

    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        return partial(self.get, key)


from typing import Any, Protocol

from pyjudo.iservice_container import IServiceContainer


class Factory[T](Protocol):
    def __call__(self, **overrides: Any) -> T: ...


class FactoryProxy[T](Factory[T]):
    _container: IServiceContainer
    _abstract_class: type[T]

    def __init__(self, container: IServiceContainer, abstract_class: type[T]):
        self._container = container
        self._abstract_class = abstract_class

    def __call__(self, **overrides: Any) -> T:
        return self._container.get(self._abstract_class, **overrides)

    def __repr__(self) -> str:
        return f"FactoryProxy({self._abstract_class.__name__})"